/**
 * Open Command Prompt Shell Extension
 * Copyright (C) 2007 Kai Liu.  All rights reserved.
 *
 * Please refer to readme.txt for information about this source code.
 * Please refer to license.txt for details about distribution and modification.
 **/

#include "globals.h"
#include "CCmdOpen.hpp"
#include "CCmdOpenClassFactory.hpp"
#include "RegHelpers.h"

// Global globals (declared as extern in globals.h)
UINT g_cRefThisDll = 0;
HMODULE g_hModThisDll = NULL;

// Local globals
LPTSTR ModuleName = NULL;

// Prototypes for the self-install/uninstall helper functions
HRESULT Install( BOOL );
HRESULT Uninstall( );


extern "C" BOOL WINAPI DllMain( HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved )
{
	switch (dwReason)
	{
		case DLL_PROCESS_ATTACH:
			g_hModThisDll = hInstance;
			if (_winmajor < 5) return(FALSE);
			break;

		case DLL_PROCESS_DETACH:
			break;

		case DLL_THREAD_ATTACH:
			break;

		case DLL_THREAD_DETACH:
			break;
	}

	return(TRUE);
}

STDAPI DllCanUnloadNow( )
{
	return((g_cRefThisDll == 0) ? S_OK : S_FALSE);
}

STDAPI DllGetClassObject( REFCLSID rclsid, REFIID riid, LPVOID *ppv )
{
	*ppv = NULL;

	if (IsEqualIID(rclsid, CLSID_CmdOpen))
	{
		LPCCMDOPENCLASSFACTORY lpCmdOpenClassFactory = new CCmdOpenClassFactory;
		if (lpCmdOpenClassFactory == NULL) return(E_OUTOFMEMORY);

		HRESULT hr = lpCmdOpenClassFactory->QueryInterface(riid, ppv);
		lpCmdOpenClassFactory->Release();
		return(hr);
	}

	return(CLASS_E_CLASSNOTAVAILABLE);
}

STDAPI DllRegisterServer( )
{
	HKEY hKey;

	// Register class
	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("CLSID\\%s"), CLSID_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSNAME_STR_CmdOpen);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("CLSID\\%s\\InprocServer32"), CLSID_STR_CmdOpen))
	{
		if (ModuleName == NULL)
		{
			// Get the module name if one is not already pre-defined.
			TCHAR szFile[MAX_PATH << 1] = { 0 };
			GetModuleFileName(g_hModThisDll, szFile, countof(szFile));
			ModuleName = szFile;
		}

		RegSetSZ(hKey, NULL, ModuleName);
		RegSetSZ(hKey, TEXT("ThreadingModel"), TEXT("Apartment"));
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	// Register handlers
	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("Directory\\Background\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
		RegSetSZ(hKey, TEXT("Type"), CLSTYPE_STR_CmdOpenBG);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("Directory\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
		RegSetSZ(hKey, TEXT("Type"), CLSTYPE_STR_CmdOpenFolder);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("Drive\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
		RegSetSZ(hKey, TEXT("Type"), CLSTYPE_STR_CmdOpenFolder);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("CLSID\\{450D8FBA-AD25-11D0-98A8-0800361B1103}\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
		RegSetSZ(hKey, TEXT("Type"), CLSTYPE_STR_CmdOpenMyDocs);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	// Register approval
	if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Shell Extensions\\Approved"), NULL))
	{
		RegSetSZ(hKey, CLSID_STR_CmdOpen, CLSNAME_STR_CmdOpen);
		RegCloseKey(hKey);
	}

	return(S_OK);
}

STDAPI DllUnregisterServer( )
{
	HKEY hKey;
	BOOL bClassRemoved = TRUE;
	BOOL bApprovalRemoved = FALSE;

	// Unregister class
	if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("CLSID\\%s"), CLSID_STR_CmdOpen))
		bClassRemoved = FALSE;

	// Unregister handlers
	if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("Directory\\Background\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
		bClassRemoved = FALSE;

	if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("Directory\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
		bClassRemoved = FALSE;

	if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("Drive\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
		bClassRemoved = FALSE;

	if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("CLSID\\{450D8FBA-AD25-11D0-98A8-0800361B1103}\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
		bClassRemoved = FALSE;

	// Unregister approval
	if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Shell Extensions\\Approved"), NULL))
	{
		LONG lResult = RegDeleteValue(hKey, CLSID_STR_CmdOpen);
		bApprovalRemoved = (lResult == ERROR_SUCCESS || lResult == ERROR_FILE_NOT_FOUND);
		RegCloseKey(hKey);
	}

	if (!bClassRemoved) return(SELFREG_E_CLASS);
	if (!bApprovalRemoved) return(S_FALSE);

	return(S_OK);
}

STDAPI DllInstall( BOOL bInstall, LPCWSTR pszCmdLine )
{
	// To install with bCopyFile=true
	// regsvr32 /i /n CmdOpen.dll
	//
	// To install with bCopyFile=false
	// regsvr32 /i:"NoCopy" /n CmdOpen.dll
	//
	// To uninstall
	// regsvr32 /u /i /n CmdOpen.dll
	//
	// DllInstall can also be invoked from a RegisterDlls INF section or from
	// a UnregisterDlls INF section, if the registration flags are set to 2.
	// Consult the documentation for RegisterDlls/UnregisterDlls for details.

	return( (bInstall) ?
		Install(pszCmdLine == NULL || lstrcmpiW(pszCmdLine, TEXT("NoCopy"))) :
		Uninstall()
	);
}

HRESULT Install( BOOL bCopyFile )
{
	// If bCopyFile=true, this function will fail if %SystemRoot%\System32\ShellExt
	// does not exist or if CmdOpen.dll is in-use and cannot be overwritten.
	// It is therefore recommended that the file copy be performed by an external
	// installer (such as an INF script) and then afterwards invoke DllInstall
	// with the "NoCopy" option.

	TCHAR szSysDir[MAX_PATH] = { 0 };
	UINT uSize = GetSystemDirectory(szSysDir, countof(szSysDir));

	TCHAR szSrc[MAX_PATH << 1] = { 0 };
	GetModuleFileName(g_hModThisDll, szSrc, countof(szSrc));

	if (uSize && uSize < countof(szSysDir))
	{
		if (szSysDir[uSize - 1] == TEXT('\\')) szSysDir[uSize - 1] = 0;

		TCHAR szDest[MAX_PATH + 0x20];
		TCHAR szCmd[MAX_PATH + 0x20];
		wsprintf(szDest, TEXT("%s\\ShellExt\\CmdOpen.dll"), szSysDir);
		wsprintf(szCmd, TEXT("%s\\cmd.exe"), szSysDir);

		// Let DllRegisterServer know which DLL it should register
		ModuleName = (bCopyFile) ? szDest : szSrc;

		if ( (!bCopyFile || lstrcmpi(szSrc, szDest) == 0 || CopyFile(szSrc, szDest, FALSE) != 0) &&
		     DllRegisterServer() == S_OK )
		{
			HKEY hKey;

			// DisableUNCCheck (KB156276)
			if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Command Processor"), NULL))
			{
				RegSetDW(hKey, TEXT("DisableUNCCheck"), 1);
				RegCloseKey(hKey);
			}

			// Uninstaller entries
			RegDelete(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s"), CLSNAME_STR_CmdOpen);

			if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s"), CLSNAME_STR_CmdOpen))
			{
				TCHAR szUninstall[MAX_PATH << 1];
				wsprintf(szUninstall, TEXT("regsvr32.exe /u /s /i /n \"%s\""), ModuleName);

				RegSetSZ(hKey, TEXT("DisplayIcon"), szCmd);
				RegSetSZ(hKey, TEXT("DisplayName"), TEXT("Open Command Prompt Shell Extension"));
				RegSetSZ(hKey, TEXT("DisplayVersion"), TEXT(CMDOPEN_VERSION_STR));
				RegSetSZ(hKey, TEXT("HelpLink"), TEXT("http://www.ktechcomputing.com/cmdopen/"));
				RegSetDW(hKey, TEXT("NoModify"), 1);
				RegSetDW(hKey, TEXT("NoRepair"), 1);
				RegSetSZ(hKey, TEXT("Publisher"), TEXT("Kai Liu"));
				RegSetSZ(hKey, TEXT("UninstallString"), szUninstall);
				RegSetSZ(hKey, TEXT("URLInfoAbout"), TEXT("http://www.ktechcomputing.com/"));
				RegCloseKey(hKey);
			}

			return(S_OK);

		} // if copied & registered

	} // if valid sysdir

	return(E_FAIL);
}

HRESULT Uninstall( )
{
	HRESULT hr = S_OK;

	// Schedule the DLL to be deleted at shutdown/reboot
	TCHAR szModule[MAX_PATH << 1] = { 0 };
	GetModuleFileName(g_hModThisDll, szModule, countof(szModule));
	if (!MoveFileEx(szModule, NULL, MOVEFILE_DELAY_UNTIL_REBOOT)) hr = E_FAIL;

	// Unregister
	if (DllUnregisterServer() != S_OK) hr = E_FAIL;

	// We don't need the uninstall strings any more...
	RegDelete(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s"), CLSNAME_STR_CmdOpen);

	return(hr);
}
